from concurrent.futures import FIRST_COMPLETED, wait, ThreadPoolExecutor
import threading
import time
from typing import Dict, List, Set, Iterable

from .object import Object
from .process import Process
from .pipeline import Pipeline

class Scheduler:
    def __init__(self, pipeline: Pipeline, processes: Iterable[Process], objects: Iterable[Object], parallelization: int = 4):
        self.pipeline: Pipeline = pipeline
        self.objects: Dict[str, Object] = {obj.name: obj for obj in objects}
        self.processes: Dict[str, Process] = {proc.name: proc for proc in processes}
        self.daemons: List[threading.Thread] = []
        self.thread_pool: ThreadPoolExecutor = ThreadPoolExecutor(max_workers=parallelization)
        self._assert_pipeline_consistent()

    def start(self):
        for obj in self.objects.values():
            obj.on_pipeline_start()
        self.start_daemons()
        
        # Main Loop
        try:
            while True:
                if self.step():
                    print("-=-=-=-=-=-=-=-=-=-")

                # Sleep to avoid CPU spin? The user didn't specify.
                # If we have run_daemons producing data, we want to pick it up fast.
                # But minimal sleep is good practice.
                time.sleep(0.01)
        except KeyboardInterrupt:
            pass

        for obj in self.objects.values():
            obj.on_pipeline_end()

    def step(self):
        """
        Runs one iteration of the topological process loop.
        """
        process_pending_inputs = {name: len(self.pipeline.process_inputs(name)) for name in self.processes}
        ready_to_poll = lambda p: process_pending_inputs[p] == 0
        poll = lambda p: self.thread_pool.submit(self._poll_process, p)

        pending_processes = set()
        for proc_name, proc in self.processes.items():
            if ready_to_poll(proc_name):
                pending_processes.add(poll(proc_name))
        while pending_processes:
            done, pending_processes = wait(pending_processes, return_when=FIRST_COMPLETED)
            for future in done:
                proc_name = future.result()
                for out_obj in self.pipeline.process_outputs(proc_name).values():
                    for consumer in self.pipeline.object_consumers(out_obj):
                        process_pending_inputs[consumer] -= 1
                        if ready_to_poll(consumer):
                            pending_processes.add(poll(consumer))

        for obj in self.objects.values():
            obj.save()

    def start_daemons(self):
        for name, proc in self.processes.items():
            if proc.has_daemon:
                # Pass inputs/outputs as kwargs?
                # "It has read access to its outputs... but we guarantee it will not change it."
                # We pass the same objects.
                kwargs = self._get_process_args(name)

                # Daemon thread
                t = threading.Thread(target=proc.run_daemon, kwargs=kwargs, name=f"daemon-{name}", daemon=True)
                t.start()
                self.daemons.append(t)
        return

    def _assert_pipeline_consistent(self):
        """
        Assert that the pipeline is consistent with the processes and objects.
        """
        assert set(self.processes.keys()) == set(self.pipeline.processes.keys()), \
            str(self.processes.keys()) + " != " + str(self.pipeline.processes.keys())
        for proc_name, proc in self.processes.items():
            assert proc_name == proc.name
            assert set(proc.inputs) == set(self.pipeline.process_inputs(proc_name).keys())
            assert set(proc.outputs) == set(self.pipeline.process_outputs(proc_name).keys())

        assert len(self.objects) == len(self.pipeline.objects)
        for obj_name, obj in self.objects.items():
            assert obj_name == obj.name
            assert obj_name in self.pipeline.objects

    def _poll_process(self, proc_name: str):
        """
        Poll a process and return the name of the process.
        """
        proc = self.processes[proc_name]
        args = self._get_process_args(proc_name)
        proc.poll(**args)

        # Return the name of the process to track process in threadpool futures
        return proc_name

    def _get_process_args(self, proc_name: str):
        kwargs = {}
        for input_port, obj_name in self.pipeline.process_inputs(proc_name).items():
            kwargs[input_port] = self.objects[obj_name]
        for output_port, obj_name in self.pipeline.process_outputs(proc_name).items():
            kwargs[output_port] = self.objects[obj_name]
        return kwargs



from lazydag.core.misc import get_processes_and_objects


def get_process_by_name(process_name: str):
    processes, _ = get_processes_and_objects()
    for proc in processes:
        if proc.name == process_name:
            return proc
    raise ValueError(f"Process {process_name} does not exist")

def get_object_by_name(object_name: str):
    _, objects = get_processes_and_objects()
    for obj in objects:
        if obj.name == object_name:
            return obj
    raise ValueError(f"Object {object_name} does not exist")



import typer
from lazydag.core.misc import get_processes_and_objects
from lazydag.core.pipeline import Pipeline
from lazydag.core.scheduler import Scheduler

run_app = typer.Typer()


@run_app.command()
def run(ctx: typer.Context):
    pipeline: Pipeline = ctx.obj["pipeline"]
    if pipeline is None:
        typer.echo("Error: pipeline not found, have you built it?")
        return
    processes, objects = get_processes_and_objects()
    scheduler = Scheduler(pipeline, processes, objects)
    scheduler.start()


if __name__ == "__main__":
    run_app()



import shutil
from pathlib import Path
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Union
import copy
import os
import pickle


class Object(ABC):
    """
    Base class for all objects.
    """

    def __init__(self, name: str):
        self.name = name

    @abstractmethod
    def on_add_to_pipeline(self):
        """
        This function is called once the object is added to the pipeline.
        """
        pass

    @abstractmethod
    def on_remove_from_pipeline(self):
        """
        This function is called once the object is removed from the pipeline.
        """
        pass

    @abstractmethod
    def on_pipeline_start(self):
        """
        This function is called once the pipeline execution starts.
        """
        pass

    @abstractmethod
    def on_pipeline_end(self):
        """
        This function is called once the pipeline execution ends.
        """
        pass

    @abstractmethod
    def save(self):
        """
        This function is called once one iteration of the pipeline is completed.
        Therefore, if there are unsaved changes, they have to be saved permanently.
        """
        pass

    @abstractmethod
    def purge(self):
        """
        This function is called when the contents of the object need to be reset
        to the initial state, i.e. the same state as after calling on_add_to_pipeline.

        An example use case is when the producer of the object is changed and needs to
        be re-executed.
        """
        pass




from typing import Dict, List, Optional, Set
from pathlib import Path
import yaml

class Pipeline:
    def __init__(self):
        self.processes = dict()
        self.objects = dict()

    @classmethod
    def from_yaml_file(cls, path: Path):
        if not path.exists():
            return None

        self = cls()
        with open(path, "r") as f:
            cfg = yaml.safe_load(f)
        for obj in cfg["objects"]:
            self.add_object(obj)
        for proc_name, proc in cfg["processes"].items():
            self.add_process(proc_name, inputs=proc.get("inputs", {}), outputs=proc.get("outputs", {}))
        return self

    def to_yaml_file(self, path: Path):
        data = {
            "objects": sorted(self.objects.keys()),
            "processes": self.processes,
        }
        with open(path, "w") as f:
            yaml.safe_dump(data, f)

    def add_object(self, name: str):
        if name in self.objects:
            raise ValueError(f"Object {name} already exists")
        self.objects[name] = {
            "producer": None,
            "consumers": set()
        }

    def add_process(self, name: str, inputs: Dict[str, str], outputs: Dict[str, str]):
        if name in self.processes:
            raise ValueError(f"Process {name} already exists")

        for inp in inputs.values():
            if inp not in self.objects:
                raise ValueError(f"Input object {inp} does not exist")
            self.objects[inp]["consumers"].add(name)

        for out in outputs.values():
            if out not in self.objects:
                raise ValueError(f"Output object {out} does not exist")
            if self.objects[out]["producer"] is not None:
                raise ValueError(f"Output object {out} is already used by process {self.objects[out]["producer"]}")
            self.objects[out]["producer"] = name

        self.processes[name] = {"inputs": inputs, "outputs": outputs}

    def remove_process(self, name: str):
        if name not in self.processes:
            raise ValueError(f"Process {name} does not exist")

        for inp in self.processes[name]["inputs"].values():
            self.objects[inp]["consumers"].remove(name)
        for out in self.processes[name]["outputs"].values():
            self.objects[out]["producer"] = None
        del self.processes[name]

    def remove_object(self, name: str):
        if name not in self.objects:
            raise ValueError(f"Object {name} does not exist")

        if self.objects[name]["producer"] is not None:
            raise ValueError(f"Object {name} is still used by process {self.objects[name]["producer"]}")
        if len(self.objects[name]["consumers"]) > 0:
            raise ValueError(f"Object {name} is still used by processes {self.objects[name]["consumers"]}")

        del self.objects[name]

    def validate(self) -> List[str]:
        errors = []
        for obj_name, obj in self.objects.items():
            if obj["producer"] is None:
                errors.append(f"Object {obj_name} is not used by any process")
            if len(obj["consumers"]) == 0:
                errors.append(f"Object {obj_name} is not used by any process")

        try:
            self.topological_sort()
        except ValueError as e:
            errors.append(str(e))

        return errors

    def topological_sort(self) -> List[str]:
        input_degree = {proc: len(self.processes[proc]["inputs"]) for proc in self.processes}
        queue = {proc for proc, degree in input_degree.items() if degree == 0}
        topological_order = []
        while len(queue) > 0:
            u = queue.pop()
            topological_order.append(u)
            for obj in self.processes[u]["outputs"].values():
                for v in self.objects[obj]["consumers"]:
                    input_degree[v] -= 1
                    if input_degree[v] == 0:
                        queue.add(v)
        if len(topological_order) != len(self.processes):
            raise ValueError("Graph contains a cycle")
        return topological_order

    def process_inputs(self, process_name: str) -> Dict[str, str]:
        return self.processes[process_name]["inputs"]

    def process_outputs(self, process_name: str) -> Dict[str, str]:
        return self.processes[process_name]["outputs"]

    def object_consumers(self, object_name: str) -> Set[str]:
        return self.objects[object_name]["consumers"]

    def object_producer(self, object_name: str) -> Optional[str]:
        return self.objects[object_name]["producer"]



from pathlib import Path
import typer

FILE_CONTENTS = {
    "settings.py": """\
from pathlib import Path

PROJECT_NAME = "{project_name}"
DATA_ROOT = Path("./data")
PY_MODULE = "defs"

FS_OBJECTS = {{
    "save_dir": DATA_ROOT / "objects",
}}
""",
    "topology.yaml": """\
processes: {{}}
objects: []
""",
    "defs.py": """\
objects = []
processes = []
""",
}

def start_project(project_name: str):
    # TODO: change the template to use some features of lazydag
    project_path = Path(project_name)
    if project_path.exists():
        raise ValueError(f"Project {project_name} already exists")

    typer.echo(f"Creating project {project_name}")
    project_path.mkdir()
    for file, content in FILE_CONTENTS.items():
        typer.echo(f"Creating file {file}")
        with open(project_path / file, "w") as f:
            f.write(content.format(project_name=project_name))



from pathlib import Path
from importlib import import_module
from types import ModuleType
import os
from typing import Any


class LazySettings:
    """
    Lazy proxy:
    - reads env var on first attribute access
    - creates a Settings instance and then delegates attribute access to it
    """
    def __init__(self, env_var: str = "LAZYDAG_SETTINGS_MODULE", default_module: str = "settings"):
        self._module: ModuleType | None = None
        self._env_var = os.getenv(env_var, default_module)

    def _setup(self) -> None:
        module_name = self._env_var
        try:
            self._module = import_module(module_name)
        except ImportError:
            raise ImportError(f"Could not import settings module: {module_name}")

    def __getattr__(self, name: str) -> Any:
        if self._module is None:
            self._setup()
        return getattr(self._module, name)

    def __setattr__(self, name: str, value: Any) -> None:
        # allow internal attrs
        if name.startswith("_"):
            return super().__setattr__(name, value)
        if self._module is None:
            self._setup()
        setattr(self._module, name, value)

    def configured(self) -> bool:
        return self._module is not None


settings = LazySettings()



from pathlib import Path
from typing import List, Dict
import typer
from lazydag.core.object import Object
from lazydag.core.pipeline import Pipeline
from lazydag.core.paths import get_pipeline_path
from lazydag.cli.utils import get_object_by_name


topology_app = typer.Typer()


@topology_app.command()
def add_process(ctx: typer.Context, process_name: str, inputs: List[str] = [], outputs: List[str] = []):
    pipeline: Pipeline = ctx.obj["pipeline"]
    inputs: Dict[str, str] = {inp.split(":")[0]: inp.split(":")[1] for inp in inputs}
    outputs: Dict[str, str] = {out.split(":")[0]: out.split(":")[1] for out in outputs}
    pipeline.add_process(process_name, inputs, outputs)
    pipeline.to_yaml_file(get_pipeline_path())


@topology_app.command()
def remove_process(ctx: typer.Context, process_name: str):
    pipeline: Pipeline = ctx.obj["pipeline"]
    pipeline.remove_process(process_name)
    pipeline.to_yaml_file(get_pipeline_path())


@topology_app.command()
def add_object(ctx: typer.Context, object_name: str):
    pipeline: Pipeline = ctx.obj["pipeline"]
    pipeline.add_object(object_name)
    pipeline.to_yaml_file(get_pipeline_path())

    object: Object = get_object_by_name(object_name)
    object.on_add_to_pipeline()


@topology_app.command()
def remove_object(ctx: typer.Context, object_name: str):
    pipeline: Pipeline = ctx.obj["pipeline"]
    pipeline.remove_object(object_name)
    pipeline.to_yaml_file(get_pipeline_path())

    object: Object = get_object_by_name(object_name)
    object.on_remove_from_pipeline()


@topology_app.command()
def from_yaml(ctx: typer.Context, yaml_addr: Path):
    if not yaml_addr.exists():
        typer.echo(f"Error: topology file {yaml_addr} does not exist")
        return

    pipeline: Pipeline = ctx.obj["pipeline"]
    if pipeline is not None:
        typer.echo("Error: pipeline already exists, remove it first")
        return

    pipeline = Pipeline.from_yaml_file(yaml_addr)
    assert pipeline is not None

    pipeline.to_yaml_file(get_pipeline_path())
    for obj_name in pipeline.objects:
        obj = get_object_by_name(obj_name)
        obj.on_add_to_pipeline()


@topology_app.command()
def validate(ctx: typer.Context):
    pipeline: Pipeline = ctx.obj["pipeline"]
    errors = pipeline.validate()
    if len(errors) > 0:
        for err in errors:
            print(err)
        raise ValueError("Pipeline is invalid")
    print("Pipeline is valid")


if __name__ == "__main__":
    topology_app()



"""
it's hard to decide on the files for these codes
I'll write them here for now
"""
import importlib
from pathlib import Path
from lazydag.conf import settings


def scaffold_data_dir(data_root: Path):
    data_root.mkdir(exist_ok=True)
    (data_root / "configs").mkdir(exist_ok=True)


def get_processes_and_objects():
    module = importlib.import_module(settings.PY_MODULE)
    return module.processes, module.objects



import copy
import os
from pathlib import Path
import pickle
import shutil
from typing import Any, List, Tuple

from lazydag.core.object import Object
from lazydag.conf import settings

class FSBackedObject(Object):
    """
    Base class for all objects stored on the filesystem.
    Manages filesystem path and abstract save/load mechanism.
    """
    def __init__(self, name: str, save_path: Path = None):
        super().__init__(name)
        self.save_path: Path = save_path or Path(settings.FS_OBJECTS["save_dir"]) / name

    def on_add_to_pipeline(self):
        self.save_path.mkdir(parents=True)

    def on_remove_from_pipeline(self):
        shutil.rmtree(self.save_path)

    def purge(self):
        for child in os.listdir(self.save_path):
            shutil.rmtree(self.save_path / child)

    def __str__(self):
        return f"{self.__class__.__name__}<{self.name}>"


class FSListObject(FSBackedObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._changelog: List[Tuple[str, int, Any]] = []

    def _get_data_path(self) -> Path:
        return self.save_path / "data.pkl"

    def _get_empty_structure(self) -> Any:
        return []

    def on_pipeline_start(self):
        data_path = self._get_data_path()
        if data_path.exists():
            with data_path.open("rb") as f:
                self._data = pickle.load(f)
        else:
            self._data = self._get_empty_structure()
        self._current = copy.deepcopy(self._data)

    def on_pipeline_end(self):
        pass

    def get(self, idx: int, old: bool = False) -> Any:
        if old:
            return self._data[idx]
        else:
            return self._current[idx]

    def insert(self, idx: int, value: Any):
        if idx < 0 or idx > len(self):
            raise ValueError("Index out of bounds")
        self._current.insert(idx, value)
        self._changelog.append(('insert', idx, value))

    def push(self, value: Any):
        """Convenience method to append to the end of the list."""
        self.insert(len(self), value)
    
    def remove(self, idx: int):
        if idx < 0 or idx >= len(self):
            raise ValueError("Index out of bounds")
        val = self._current.pop(idx)
        self._changelog.append(('remove', idx, val))

    def set(self, idx: int, value: Any):
        if idx < 0 or idx >= len(self):
            raise ValueError("Index out of bounds")
        if self._current[idx] == value:
            return
        self._current[idx] = value
        self._changelog.append(('set', idx, value))

    def save(self):
        # Update persistent data
        self._data = copy.deepcopy(self._current)

        data_path = self._get_data_path()
        with data_path.open("wb") as f:
            pickle.dump(self._data, f)

        # Clear change history
        self._changelog.clear()

    def changed(self) -> bool:
        return len(self._changelog) > 0

    def __len__(self):
        return len(self._current)

    def __iter__(self):
        return iter(self._current)

    def __getitem__(self, idx: int):
        return self._current[idx]


class FSDictObject(FSBackedObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _get_data_path(self) -> Path:
        return self.save_path / "data.pkl"

    def _get_empty_structure(self) -> Any:
        return {}

    def on_pipeline_start(self):
        data_path = self._get_data_path()
        if data_path.exists():
            with data_path.open("rb") as f:
                self._data = pickle.load(f)
        else:
            self._data = self._get_empty_structure()
        self._current = copy.deepcopy(self._data)

    def on_pipeline_end(self):
        pass

    def get(self, key: Any, old: bool = False) -> Any:
        if old:
            return self._data.get(key)
        else:
            return self._current.get(key)

    def set(self, key: Any, value: Any):
        self._current[key] = value
        self._changelog.append(('set', key, value))

    def remove(self, key: Any):
        if key in self._current:
            del self._current[key]
            self._changelog.append(('remove', key))

    def save(self):
        self._data = copy.deepcopy(self._current)

        os.makedirs(os.path.dirname(self.save_path), exist_ok=True)

        with open(self.save_path, 'wb') as f:
            pickle.dump(self._data, f)

        self._changelog.clear()

    def changed(self) -> bool:
        return len(self._changelog) > 0

    def __len__(self):
        return len(self._current)

    def __iter__(self):
        return iter(self._current)

    def __contains__(self, key: Any):
        return key in self._current

    def keys(self):
        return self._current.keys()

    def values(self):
        return self._current.values()

    def items(self):
        return self._current.items()

    def __getitem__(self, key: Any):
        return self._current[key]



import os
import sys
import typer

from lazydag.conf import settings
from lazydag.core.misc import scaffold_data_dir
from lazydag.core.pipeline import Pipeline
from lazydag.core.paths import get_pipeline_path

from lazydag.cli.topology import topology_app
from lazydag.cli.run import run_app
from lazydag.cli.appless_commands import start_project


def main():
    app = typer.Typer()
    app.add_typer(topology_app, name="topology", callback=callback)
    app.add_typer(run_app, name="run", callback=callback)
    app.command()(start_project)
    app()


def callback(ctx: typer.Context):
    sys.path.insert(0, os.getcwd())

    scaffold_data_dir(settings.DATA_ROOT)
    pipeline = Pipeline.from_yaml_file(get_pipeline_path())

    ctx.ensure_object(dict)
    ctx.obj["pipeline"] = pipeline


if __name__ == "__main__":
    main()



from abc import ABC, abstractmethod
from typing import List

class Process(ABC):
    """
    Base class for processes.
    """
    # These should be defined by subclasses
    inputs: List[str] = []
    outputs: List[str] = []
    has_daemon: bool = False

    def __init__(self, name: str):
        self.name = name

    @abstractmethod
    def poll(self, **kwargs):
        """
        Execution logic. kwargs will contain arguments matching inputs and outputs keys.
        Triggered when inputs change.
        """
        pass

    def run_daemon(self, **kwargs):
        """
        Optional method for daemon processes.
        Run in a background thread. Has read access to outputs but should not modify them directly.
        """
        pass
